import datetime
import traceback
import uuid
from logging import Logger

import asyncio
from typing import Type

from sqlalchemy import select, Sequence, and_, func, update, case, or_
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.tasks.constants import StatusTaskRosemary, TypeTaskRosemary
from rosemary.worker.constants import StatusWorkerRosemary
from rosemary.core.custom_semaphore import CustomSemaphore
from rosemary.db.db import DBConnector
from rosemary.db.models import RosemaryWorkerModel, RosemaryTaskModel
from rosemary.core.logger import get_logger
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class RosemaryWorker:
    def __init__(
            self,
            db_host: str,
            db_port: int | str,
            db_user: str,
            db_password: str,
            db_name_db: str,
            db_schema: str,
            tasks: dict[str, Type[InterfaceRosemaryTask]],
            shutdown_event,
            logger: Logger | None = None,
            max_task_semaphore: int = 30,
    ):
        self.uuid = uuid.uuid4()
        self.worker_db: RosemaryWorkerModel | None = None
        self.db_connector = None
        self.logger: Logger | None = get_logger(str(self.uuid)) or logger

        self.db_connector = DBConnector(
            db_host, db_name_db, db_user, db_password, db_port, db_schema
        )
        self._max_task_semaphore = max_task_semaphore
        self._registered_tasks = tasks
        self.__shutdown_event = shutdown_event

    def get_task_by_name(self, task_name: str) -> Type[InterfaceRosemaryTask]:
        return self._registered_tasks[task_name]

    async def _get_new_tasks(
            self, session: AsyncSession, limit: int
    ) -> Sequence[int]:
        select_query = select(RosemaryTaskModel.id).where(
            and_(
                RosemaryTaskModel.status.in_([
                    StatusTaskRosemary.NEW.value,
                    StatusTaskRosemary.FAILED.value
                ]),
                RosemaryTaskModel.delay <= func.now(),
                RosemaryTaskModel.worker.in_([
                    self.worker_db.id, None
                ])
            )
        ).limit(limit).with_for_update(skip_locked=True)

        res = await session.execute(select_query)
        ids_to_update = res.scalars().all()
        update_status = update(RosemaryTaskModel).where(
            RosemaryTaskModel.id.in_(ids_to_update)
        ).values(
            status=StatusTaskRosemary.IN_PROGRESS.value,
            worker=self.worker_db.id,
            retry=case(
                (RosemaryTaskModel.status.in_(
                    [StatusTaskRosemary.FAILED.value, StatusTaskRosemary.IN_PROGRESS.value]
                ), RosemaryTaskModel.retry + 1),  # condition and result as a tuple
                else_=RosemaryTaskModel.retry
            )
        )
        await session.execute(update_status)
        await session.commit()
        return ids_to_update

    async def register_in_db(self, session: AsyncSession):
        self.worker_db = RosemaryWorkerModel(uuid=self.uuid)
        session.add(self.worker_db)
        await session.commit()

    async def __ping(self, session: AsyncSession):
        self.worker_db.ping_time = func.now()
        self.worker_db.status = StatusWorkerRosemary.WORKING.value
        await session.commit()

    async def __suicide(self, session: AsyncSession):
        self.worker_db.status = StatusWorkerRosemary.KILLED.value
        await session.commit()

    def run(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self._looping())
        loop.close()

    async def _looping(self):
        step = 60 * 60
        async with self.db_connector.get_session() as session:
            await self.register_in_db(session)
            self.logger.info(f'Start looping by worker {self.uuid}')
            semaphore = CustomSemaphore(self._max_task_semaphore)
            while not self.__shutdown_event.is_set():
                await self.__ping(session)
                if semaphore.tasks_remaining() > 0:
                    ids_tasks = await self._get_new_tasks(session, semaphore.tasks_remaining())
                    for id_task in ids_tasks:
                        await semaphore.acquire()
                        asyncio.create_task(self._run_task(id_task, semaphore))
                else:
                    await asyncio.sleep(2)
                await asyncio.sleep(1)

                step += 1

                if step >= 60 * 1:
                    await self.__check_stuck_tasks(session)

                if step >= 60 * 2:
                    step = 0
                    await self._check_deaths_workers(session)

            self.logger.info(f'Rosemary worker {self.uuid} is shutdowning warm...')
            while semaphore.tasks_remaining() != self._max_task_semaphore:
                await asyncio.sleep(1)
            await self.__suicide(session)
            self.logger.info(f'Rosemary worker {self.uuid} is shutdown warm!')

    async def __check_stuck_tasks(self, session: AsyncSession):
        select_query = select(RosemaryTaskModel.id).join(RosemaryWorkerModel).where(
            and_(
                RosemaryTaskModel.status == StatusTaskRosemary.IN_PROGRESS.value,
                RosemaryTaskModel.worker == self.worker_db.id,
                or_(
                    RosemaryTaskModel.updated_at < (
                            func.now() - func.make_interval(0, 0, 0, 0, 0, 0, (RosemaryTaskModel.timeout + 10))
                    ),
                    RosemaryWorkerModel.status == StatusWorkerRosemary.KILLED.value,
                ),
                (RosemaryTaskModel.retry + 1) >= RosemaryTaskModel.max_retry
            )
        ).with_for_update(skip_locked=True)
        res = await session.execute(select_query)
        ids_to_update = res.scalars().all()
        update_status = update(RosemaryTaskModel).where(
            RosemaryTaskModel.id.in_(ids_to_update)
        ).values(
            status=StatusTaskRosemary.FATAL.value,
            retry=RosemaryTaskModel.retry + 1
        )
        await session.execute(update_status)
        await session.commit()

        select_query = select(RosemaryTaskModel.id).join(RosemaryWorkerModel).where(
            and_(
                RosemaryTaskModel.status == StatusTaskRosemary.IN_PROGRESS.value,
                RosemaryTaskModel.worker == self.worker_db.id,
                or_(
                    RosemaryTaskModel.updated_at < (
                            func.now() - func.make_interval(0, 0, 0, 0, 0, 0, (RosemaryTaskModel.timeout + 10))
                    ),
                    RosemaryWorkerModel.status == StatusWorkerRosemary.KILLED.value,
                ),
                (RosemaryTaskModel.retry + 1) < RosemaryTaskModel.max_retry
            )
        ).with_for_update(skip_locked=True)
        res = await session.execute(select_query)
        ids_to_update = res.scalars().all()
        update_status = update(RosemaryTaskModel).where(
            RosemaryTaskModel.id.in_(ids_to_update)
        ).values(
            status=StatusTaskRosemary.FAILED.value,
            retry=RosemaryTaskModel.retry + 1
        )
        await session.execute(update_status)
        await session.commit()

    async def _check_deaths_workers(self, session: AsyncSession):
        select_query = select(RosemaryWorkerModel.id).where(
            and_(
                RosemaryWorkerModel.status == StatusWorkerRosemary.WORKING.value,
                RosemaryWorkerModel.updated_at < (
                        func.now() - func.make_interval(0, 0, 0, 0, 0, 10, 0)
                ),
            )
        ).with_for_update(skip_locked=True)
        res = await session.execute(select_query)
        ids_to_update = res.scalars().all()
        if ids_to_update:
            self.logger.error(f'Workers {ids_to_update} is checking')
            update_status = update(RosemaryWorkerModel).where(
                RosemaryWorkerModel.id.in_(ids_to_update)
            ).values(
                status=StatusWorkerRosemary.CHECKING.value
            )
            await session.execute(update_status)
        await session.commit()
        select_query = select(RosemaryWorkerModel.id).where(
            and_(
                RosemaryWorkerModel.status == StatusWorkerRosemary.CHECKING.value,
                RosemaryWorkerModel.updated_at < (
                        func.now() - func.make_interval(0, 0, 0, 0, 0, 10, 0)
                ),
            )
        ).with_for_update(skip_locked=True)
        res = await session.execute(select_query)
        ids_to_update = res.scalars().all()
        if ids_to_update:
            self.logger.error(f'Workers {ids_to_update} is killed!!!')
            update_status = update(RosemaryWorkerModel).where(
                RosemaryWorkerModel.id.in_(ids_to_update)
            ).values(
                status=StatusWorkerRosemary.KILLED.value
            )
            await session.execute(update_status)

    async def _run_task(self, id_task: int, pool: CustomSemaphore):
        error: str | None = None
        task: InterfaceRosemaryTask | None = None

        try:
            async with self.db_connector.get_session() as session:
                try:
                    query = select(RosemaryTaskModel).where(
                        RosemaryTaskModel.id == id_task
                    )
                    result = await session.execute(query)
                    task_db: RosemaryTaskModel = result.scalars().one()
                except Exception as e:
                    self.logger.exception(f'Error while getting task {id_task} from DB {e}', exc_info=e)
                    return
                try:
                    task = self.get_task_by_name(task_db.name)()
                except KeyError:
                    error = 'Task not registered in rosemary!'
                except Exception as e:
                    error = f'{e.__class__.__name__}: {repr(e)}. Traceback: {traceback.print_tb(e.__traceback__)}'
                else:
                    try:
                        self.logger.info(f'Start task "{task_db.name}" id: "{task_db.id}" with data {task_db.data}')
                        result_task = await asyncio.wait_for(
                            task.prepare_and_run(task_db.data, session), timeout=task_db.timeout
                        )
                        self.logger.info(f'Finished task "{task_db.name}" id: "{task_db.id}" '
                                         f'with data {task_db.data} with result: {result_task}')
                    except Exception as e:
                        if isinstance(e, asyncio.TimeoutError):
                            error = f'TimeoutError: The task has timed out {task_db.timeout}'
                        else:
                            error = f'{e.__class__.__name__}: {repr(e)}. Traceback: {traceback.print_tb(e.__traceback__)}'
                        self.logger.info(f'Error task "{task_db.name}" id: "{task_db.id}" '
                                         f'with data {task_db.data} with error: {error}')
                if error:
                    task_db.error = error
                    if task_db.retry >= task_db.max_retry:
                        will_not_repeat = True
                        task_db.status = StatusTaskRosemary.FATAL.value
                    else:
                        will_not_repeat = False
                        task_db.status = StatusTaskRosemary.FAILED.value
                        task_db.delay = task.delay_retry.get_datetime_plus_interval()
                else:
                    task_db.status = StatusTaskRosemary.FINISHED.value
                    task_db.task_return = str(result_task)
                    will_not_repeat = True
                await session.commit()
            if (will_not_repeat
                    and task and task.get_type() == TypeTaskRosemary.REPEATABLE.value
                    and self.get_task_by_name(task_db.name)):
                await task.create(
                    data=task_db.data, session=session, delay=task.delay.get_datetime_plus_interval()
                )
        except Exception as e:
            self.logger.exception(f'Error while creating session for DB {e}. Task: {id_task}', exc_info=e)
        finally:
            await pool.release()



from pathlib import Path

from alembic.config import Config
from alembic import command
import logging


class Alembic:
    def __init__(self, host: str, db: str, user: str, password: str | None, port: int | str, schema: str):
        alembic_logger = logging.getLogger('alembic')
        handler = logging.StreamHandler()  # Вы можете выбрать другой handler при необходимости
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        alembic_logger.addHandler(handler)
        alembic_logger.setLevel(logging.DEBUG)  # Или другой уровень логирования по вашему выбору

        __connect_string = f'postgresql://{user}:{password}@{host}:{port}/{db}?options=-csearch_path={schema}'
        __root_dir = Path(__file__).parent.parent.parent
        __alembic_dir = __root_dir / 'rosemary/alembic'
        alembic_logger.info(__alembic_dir)

        self.__alembic_cfg = Config()
        self.__alembic_cfg.set_main_option("sqlalchemy.url", __connect_string)
        self.__alembic_cfg.set_main_option("script_location", str(__alembic_dir))


    def upgrade(self, revision):
        command.upgrade(self.__alembic_cfg, revision=revision)



import contextlib

from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession


class DBConnector:
    def __init__(
            self, host: str, db: str, user: str, password: str, port: int | str, schema: str
    ):
        self.__connect_string = self._build_asyncpg_connect_string(host, db, user, password, port)
        self.engine = create_async_engine(
            self.__connect_string,
            connect_args={"server_settings": {"search_path": schema}}
        )

    def _build_asyncpg_connect_string(self,
            host: str, db: str, user: str, password: str, port: int | str = 5432
    ):
        return f'postgresql+asyncpg://{user}:{password}@{host}:{port}/{db}'

    @contextlib.asynccontextmanager
    async def get_session(self):
        async with AsyncSession(self.engine, expire_on_commit=False) as session:
            yield session



import enum


class TypeTaskRosemary(enum.Enum):
    REPEATABLE = "REPEATABLE"
    CRON = "CRON"
    MANUAL = "MANUAL"
    NOT_SETUP = "NOT_SETUP"

    @classmethod
    def values(cls):
        return [member.value for member in cls]


class StatusTaskRosemary(enum.Enum):
    NEW = "NEW"
    IN_PROGRESS = "IN_PROGRESS"
    FAILED = "FAILED"
    FATAL = "FATAL"
    CANCELED = "CANCELED"
    FINISHED = 'FINISHED'



import asyncio
import signal
import threading
import time
from abc import ABC
from logging import Logger
from threading import Thread
from typing import Type, Iterable

from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.db.alembic import Alembic
from rosemary.db.db import DBConnector
from rosemary.core.logger import get_logger
from rosemary.settings import ALEMBIC_REVISION
from rosemary.tasks.constants import TypeTaskRosemary
from rosemary.worker.worker import RosemaryWorker
from rosemary.tasks.manual_task import InterfaceManualTask
from rosemary.tasks.repeatable_task import InterfaceRepeatableTask
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class Rosemary:

    @property
    def RepeatableTask(self) -> Type[InterfaceRepeatableTask]:
        session = self.db_connector.get_session

        class RepeatableTask(InterfaceRepeatableTask, ABC):
            @classmethod
            def get_session(cls) -> AsyncSession:
                return session()

            @classmethod
            async def run(self, data):
                ...

        return RepeatableTask

    @property
    def ManualTask(self) -> Type[InterfaceManualTask]:
        session = self.db_connector.get_session

        class ManualTask(InterfaceManualTask, ABC):
            @classmethod
            def get_session(cls) -> AsyncSession:
                return session()

        return ManualTask

    def __init__(
            self,
            *,
            db_host: str,
            db_port: int | str = 5432,
            db_user: str,
            db_password: str,
            db_name_db: str,
            db_schema: str = 'public',
            logger: Logger | None = None,
            max_tasks_per_worker: int = 50,
            workers: int = 1,
    ):
        self._max_task_semaphore: int = max_tasks_per_worker
        self.logger: Logger = logger or get_logger('Main')
        self._count_workers: int = workers

        self._workers = []
        self._registered_tasks = {}
        self._repeatable_tasks = []

        self.__db_host = db_host
        self.__db_port = db_port
        self.__db_user = db_user
        self.__db_password = db_password
        self.__db_name = db_name_db
        self.__db_schema = db_schema

        self.db_connector = DBConnector(
            host=self.__db_host,
            db=self.__db_name,
            user=self.__db_user,
            password=self.__db_password,
            port=self.__db_port,
            schema=self.__db_schema,
        )
        self.__register_signal()
        self.__shutdown_requested = False
        self.__shutdown_event = threading.Event()

    def __handle_signal(self, *args, **kwargs):
        self.logger.error(f"Rosemary is shutdowning warm...")
        self.__shutdown_requested = True
        self.__shutdown_event.set()

    def __register_signal(self):
        signal.signal(signal.SIGTERM, self.__handle_signal)
        signal.signal(signal.SIGINT, self.__handle_signal)

    def register_task(self, task: Type[InterfaceRosemaryTask]):
        ex_task = task()
        self.logger.info(f'Registered {ex_task.get_type()} task "{ex_task.get_name()}"')
        self._registered_tasks[ex_task.get_name()] = task

        if ex_task.get_type() == TypeTaskRosemary.REPEATABLE.value:
            self._repeatable_tasks.append(task)

    def _run_migration(self):
        self.logger.info('Migration is running...')
        alembic = Alembic(
            host=self.__db_host,
            db=self.__db_name,
            user=self.__db_user,
            password=self.__db_password,
            port=self.__db_port,
            schema=self.__db_schema,
        )
        alembic.upgrade(ALEMBIC_REVISION)
        self.logger.info('Migration finished!')

    @classmethod
    async def _create_tasks(cls, tasks: Iterable[Type[InterfaceRepeatableTask]]):
        await asyncio.gather(*tasks)

    def _create_worker(self, shutdown_event) -> tuple[RosemaryWorker, Thread]:
        worker = RosemaryWorker(
            db_user=self.__db_user,
            db_password=self.__db_password,
            db_name_db=self.__db_name,
            db_host=self.__db_host,
            db_schema=self.__db_schema,
            db_port=self.__db_port,
            max_task_semaphore=self._max_task_semaphore,
            tasks=self._registered_tasks,
            shutdown_event=shutdown_event,
        )
        worker.logger.info(f'Worker {worker.uuid} created!')
        thread = threading.Thread(target=worker.run)
        return worker, thread

    def run(self, repeat_tasks: Iterable[Type[InterfaceRepeatableTask]] | None = None):
        workers_threads = {}
        self._run_migration()
        if repeat_tasks:
            self.logger.info('Registration repeatable tasks')
            asyncio.run(self._create_tasks(repeat_tasks))
            self.logger.info('Registration repeatable tasks completed')

        while not self.__shutdown_requested:
            if len(workers_threads) < self._count_workers:
                for _ in range(self._count_workers):
                    worker, thread = self._create_worker(shutdown_event=self.__shutdown_event)
                    workers_threads[thread] = worker
                    thread.start()
            for_delete = []
            for th, worker in workers_threads.items():
                if not th.is_alive():
                    self.logger.error(f'Worker {worker.uuid} is deleted!')
                    for_delete.append(th)
            for th in for_delete:
                workers_threads.pop(th)
            time.sleep(4)
        self.logger.error('Rosemary is shutdown!')



import datetime
from abc import ABC

from sqlalchemy import func
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.db.models import RosemaryTaskModel
from rosemary.tasks.constants import TypeTaskRosemary
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class InterfaceManualTask(InterfaceRosemaryTask, ABC):
    max_retry = 3
    type_task = TypeTaskRosemary.MANUAL
    timeout = 30

    async def _create_task(self, data: dict, session: AsyncSession, delay: datetime.datetime):
        new_task = RosemaryTaskModel(
            data=data,
            name=self.__class__.__name__,
            type_task=self.type_task,
            max_retry=self.max_retry,
            timeout=self.timeout,
            delay=func.now() if delay is None else delay
        )
        session.add(new_task)
        await session.commit()
        return new_task.id



from datetime import datetime
import json
from abc import ABC

from sqlalchemy import select, and_, String, cast, Sequence, func
from sqlalchemy.ext.asyncio import AsyncSession

from rosemary.db.models import RosemaryTaskModel
from rosemary.tasks.constants import TypeTaskRosemary, StatusTaskRosemary
from rosemary.tasks.task_interface import InterfaceRosemaryTask


class InterfaceRepeatableTask(InterfaceRosemaryTask, ABC):
    max_retry = 0
    type_task = TypeTaskRosemary.REPEATABLE

    async def _create_task(
            self,
            data: dict,
            session: AsyncSession,
            delay: datetime
    ):
        try:
            query = select(RosemaryTaskModel).where(and_(
                RosemaryTaskModel.name == self.get_name(),
                cast(RosemaryTaskModel.data, String) == json.dumps(data)
            )).where(
                RosemaryTaskModel.status.in_([
                    StatusTaskRosemary.IN_PROGRESS.value,
                    StatusTaskRosemary.FAILED.value,
                    StatusTaskRosemary.NEW.value,
                ])
            ).with_for_update()

            result = await session.execute(query)
            task_db: Sequence[RosemaryTaskModel] = result.scalars().all()

            if task_db:
                return task_db[-1].id
        except :
            return None

        new_task = RosemaryTaskModel(
            data=data,
            name=self.__class__.__name__,
            type_task=self.type_task,
            max_retry=self.max_retry,
            timeout=self.timeout,
            delay=func.now() if delay is None else delay
        )
        session.add(new_task)
        await session.commit()
        return new_task.id



from sqlalchemy import func


class RosemaryInterval:

    def __init__(
            self,
            seconds: int = 0,
            minutes: int = 0,
            hours: int = 0,
    ):
        self.seconds = seconds
        self.minutes = minutes
        self.hours = hours

    def get_interval(self) -> int:
        return func.make_interval(0, 0, 0, 0, self.hours, self.minutes, self.seconds)

    def get_datetime_plus_interval(self):
        return self.get_interval() + func.now()



import asyncio


class CustomSemaphore(asyncio.Semaphore):
    def __init__(self, value):
        super().__init__(value)
        self._initial_value = value
        self._current_value = value

    def tasks_remaining(self):
        return self._current_value

    async def acquire(self):
        await super().acquire()
        self._current_value -= 1

    async def release(self):
        super().release()
        self._current_value += 1



"""update key to worker

Revision ID: 9bb6f1128279
Revises: 25ed0f563bdc
Create Date: 2023-12-19 10:22:34.899234

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '9bb6f1128279'
down_revision: Union[str, None] = '25ed0f563bdc'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('rosemary_tasks', 'worker')
    op.add_column('rosemary_tasks', sa.Column('worker', sa.Integer(), nullable=True))
    op.create_foreign_key(None, 'rosemary_tasks', 'rosemary_worker', ['worker'], ['id'])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'rosemary_tasks', type_='foreignkey')
    op.drop_column('rosemary_tasks', 'worker')
    op.add_column('rosemary_tasks', sa.Column('worker', sa.UUID(), nullable=True))
    # ### end Alembic commands ###



import asyncio
import datetime
import inspect
from abc import abstractmethod, ABC

from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModel_V1
from sqlalchemy.ext.asyncio import AsyncSession
from rosemary.interval import RosemaryInterval
from rosemary.tasks.constants import TypeTaskRosemary


class InterfaceRosemaryTask(ABC):
    max_retry = 3
    delay_retry: RosemaryInterval = RosemaryInterval(seconds=5)
    delay: RosemaryInterval = RosemaryInterval(seconds=10)
    type_task = TypeTaskRosemary.NOT_SETUP
    timeout = 30

    @classmethod
    def _is_event_loop_running(cls):
        try:
            asyncio.get_running_loop()
            return True
        except RuntimeError:
            return False

    def cron_setup(self, *, hour: int | str ='*', minutes: int | str = '*', days: str = '*'):
        ...

    def get_name(self) -> str:
        return self.__class__.__name__

    def get_type(self) -> str:
        return self.type_task if isinstance(self.type_task, str) else self.type_task.value

    @abstractmethod
    async def _create_task(self, data: dict, session: AsyncSession, delay: datetime.datetime) -> int:
        ...

    def prepare_data_for_run(self, data: dict):
        return data

    def _prepare_data_to_db(self, data) -> dict | None:
        if data is None:
            return data
        if isinstance(data, dict):
            return data
        if isinstance(data, BaseModel):
            return data.model_dump()
        if isinstance(data, BaseModel_V1):
            return data.dict()
        raise TypeError(f'Incorrect type of data: "{type(data)}"')

    async def create(
            self, *, data: dict | BaseModel | None = None,
            session: AsyncSession | None = None,
            delay: datetime.datetime | None = None
    ):
        data = self._prepare_data_to_db(data)

        if session is None:
            async with self.get_session() as session:
                return await self._create_task(data, session, delay)
        else:
            return await self._create_task(data, session, delay)

    async def prepare_and_run(self, data: dict | None, session: AsyncSession | None):
        params = inspect.signature(self.run)
        kwargs = {}
        for param in params.parameters.keys():
            if param == 'data':
                prepared_data = self.prepare_data_for_run(data)
                kwargs['data'] = prepared_data
            if param == 'session':
                kwargs['session'] = session
        return await self.run(**kwargs)

    @abstractmethod
    def get_session(self) -> AsyncSession:
        ...

    @abstractmethod
    async def run(self, data):
        ...
